/***************************************************************************
 *                                                                         *
 *   KCPULoad and KNetLoad are copyright (c) 1999-2000, Markus Gustavsson  *
 *                                       (c) 2002, Ben Burton              *
 *                                       (c) 2004, Diego Pettenò           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "statgraph.h"

#include <qpainter.h>
#include <qlabel.h>
#include <kdebug.h>

#define SOFT_STEP 3

const QColor StatGraph::defaultBgColor = Qt::black;
const QColor StatGraph::defaultUpperColor = Qt::red;
const QColor StatGraph::defaultLowerColor = Qt::darkRed;
const QColor StatGraph::defaultLabelColor = Qt::white;
const QColor StatGraph::defaultGridColor = Qt::lightGray;
const QColor StatGraph::invalidColor = QColor(-1, -1, -1);	///< Color used to pass transparent colors

StatGraph::StatGraph(QLabel *parent, ushort pace, ushort gridPace, const QString &label, const QColor &bgColor, const char *name)
	: QObject(parent, name), m_graph(parent), m_style(Shades), m_gridPace(gridPace)
{
	kdDebug() << "StatGraph::StatGraph(" << parent << ", " << pace << ", " << gridPace << ", " << label << ", " << bgColor << ", " << name << ")" << endl;
	setLabel(label);
	setBgColor(bgColor);
	
	m_readings = m_graph->width() / pace;
	m_currReading = 0;
	m_readingsUpper = new uint[m_readings]; memset(m_readingsUpper, 0, m_readings*sizeof(uint));
	m_readingsLower = new uint[m_readings]; memset(m_readingsLower, 0, m_readings*sizeof(uint));
	
}

StatGraph::~StatGraph()
{
	delete[] m_readingsUpper;
	delete[] m_readingsLower;
}

void StatGraph::reparent(QLabel *newParent, ushort pace)
{
	kdDebug() << "StatGraph::reparent(" << newParent << ", " << pace << ")" << endl;
	if ( m_graph != newParent )
	{
		m_graph->removeChild(this);
		newParent->insertChild(this);
		m_graph = newParent;
	}
	
	uint oldReadings = m_readings;
	m_readings = m_graph->width() / pace;
	
	if ( m_readings != oldReadings )
	{
		uint *oldReadingsUpper = m_readingsUpper;
		uint *oldReadingsLower = m_readingsLower;
		
		
		m_readingsUpper = new uint[m_readings]; memset(m_readingsUpper, 0, m_readings*sizeof(uint));
		m_readingsLower = new uint[m_readings]; memset(m_readingsLower, 0, m_readings*sizeof(uint));
	
		if ( m_readings < oldReadings )
		{
			memcpy(m_readingsUpper, oldReadingsUpper, m_readings*sizeof(uint));
			memcpy(m_readingsLower, oldReadingsLower, m_readings*sizeof(uint));
		} else {
			memcpy(m_readingsUpper, oldReadingsUpper, oldReadings*sizeof(uint));
			memcpy(m_readingsLower, oldReadingsLower, oldReadings*sizeof(uint));
		}
		delete oldReadingsUpper; delete oldReadingsLower;
		m_currReading %= m_readings;
	}
}

void StatGraph::setBgColor(const QColor &color)
{
	m_bgColor = color;

	if ( m_bgColor.isValid() )
		m_graph->setBackgroundColor(m_bgColor);
}

void StatGraph::redraw()
{
	kdDebug() << "StatGraph::redraw()" << endl;
	if ( ! m_upperColor.isValid() ) m_upperColor = defaultUpperColor;
	if ( ! m_lowerColor.isValid() ) m_lowerColor = defaultLowerColor;
	if ( ! m_labelColor.isValid() ) m_labelColor = defaultLabelColor;
	if ( ! m_gridColor.isValid() ) m_gridColor = defaultGridColor;
	
	QPainter p(m_graph);
	if ( m_gridPace )
		paintGrid(p);
	
	switch(m_style)
	{
	case Shades:
		paintShades(p); break;
	case Bars:
		paintBars(p); break;
	case Lines:
		paintLines(p); break;
	default:
		kdDebug() << "Unknown style " << m_style << endl;
	}

	if ( ! m_label.isNull() )	
		paintLabel(p);
}

inline void StatGraph::paintGrid(QPainter &p)
{
	kdDebug() << "StatGraph::paintGrid()" << endl;
	p.setPen(m_gridColor);
	
	const uint gridLines = m_graph->height() / m_gridPace;
	static uint graphWidth = m_graph->width();
	for(uint i = 0; i < gridLines; i++)
	{
		const int lineHeight = i*m_gridPace;
		p.drawLine(0, lineHeight, graphWidth-1, lineHeight);
	}
}

inline void StatGraph::paintShades(QPainter &p)
{
	kdDebug() << "StatGraph::paintShades()" << endl;
	const uint graphHeight = m_graph->height();
        for(uint i = 0; i < m_readings; i++)
	{
		ushort tmpPos = (m_currReading + i + 1) % m_readings;
		for(uint j = 0; j <= m_readingsUpper[tmpPos]; j++)
		{
			if (m_readingsUpper[tmpPos] == 0 || j == 0)
				p.setPen(m_bgColor);
			else
				p.setPen(m_upperColor.dark((100 * m_readingsUpper[tmpPos]) / j));

			p.drawPoint(i, graphHeight - 1 - j - m_readingsLower[tmpPos]);
		}

		for(uint j = 0; j <= m_readingsLower[tmpPos]; j++)
		{
			if (m_readingsLower[tmpPos] == 0 || j == 0)
				p.setPen(m_bgColor);
			else
				p.setPen(m_lowerColor.dark((100 * m_readingsLower[tmpPos]) / j));
			
			p.drawPoint(i, graphHeight - j);
		}
	}
}

inline void StatGraph::paintBars(QPainter &p)
{
	kdDebug() << "StatGraph::paintBars()" << endl;
	const uint graphHeight = m_graph->height();
	// Draw the upper bars, then the lower to save on pen adjustments.
	p.setPen(m_upperColor);
	for(uint i = 0; i < m_readings; i++)
	{
		ushort tmpPos = (m_currReading + i + 1) % m_readings;
		p.drawLine(i, graphHeight - 1 - m_readingsUpper[tmpPos] - m_readingsLower[tmpPos],
			i, graphHeight - 1);
	}
	p.setPen(m_lowerColor);
	for(uint i = 0; i < m_readings; i++)
	{
		ushort tmpPos = (m_currReading + i + 1) % m_readings;
		p.drawLine(i, graphHeight - 1 - m_readingsLower[tmpPos],
			i, graphHeight - 1);
	}
}

inline void StatGraph::paintLines(QPainter &p)
{
	kdDebug() << "StatGraph::paintLines()" << endl;
	const uint graphHeight = m_graph->height();
	// Draw the upper line, then the lower to save on pen adjustments.
	p.setPen(m_upperColor);
	for(uint i = 0; i < m_readings; i++)
	{
		ushort tmpPos = (m_currReading + i + 1) % m_readings;
		p.drawPoint(i, graphHeight - 1 -
			m_readingsUpper[tmpPos] - m_readingsLower[tmpPos]);
	}
	p.setPen(m_lowerColor);
	for(uint i = 0; i < m_readings; i++)
	{
		ushort tmpPos = (m_currReading + i + 1) % m_readings;
		p.drawPoint(i, graphHeight - 1 - m_readingsLower[tmpPos]);
	}
}

inline void StatGraph::paintLabel(QPainter &p)
{
	kdDebug() << "StatGraph::paintLabel()" << endl;
	p.setFont(QFont("Helvetica", 8));
	p.setPen(m_labelColor);
	p.drawText(m_graph->rect(), AlignLeft | AlignTop, m_label);
}

void StatGraph::clear()
{
	memset(m_readingsUpper, 0, m_readings*sizeof(uchar));
	memset(m_readingsLower, 0, m_readings*sizeof(uchar));
	m_graph->update();
}

void StatGraph::addPercentReading(uchar upper, uchar lower, bool soft)
{
	kdDebug() << "StatGraph::addPercentReading(" << (uint)upper << ", " << (uint)lower << ", " << soft << ")" << endl;
	// Rescale the readings to a measure in pixels.
	uint pixUpper = ( float(upper) * float(m_graph->height()) )/100;
	uint pixLower = ( float(lower) * float(m_graph->height()) )/100;
	
	if ( soft )
		softenReadings(pixUpper, pixLower);
	
	m_currReading = (m_currReading+1) % m_readings;
	m_readingsUpper[m_currReading] = pixUpper;
	m_readingsLower[m_currReading] = pixLower;
	m_graph->update();
}

void StatGraph::softenReadings(uint &upper, uint &lower)
{
	uint oldUpper = m_readingsUpper[m_currReading];
	uint oldLower = m_readingsLower[m_currReading];
	
	// Modify the upper reading.
	if (upper > oldUpper + SOFT_STEP)
		upper = oldUpper + SOFT_STEP;
	else if (upper < oldUpper - SOFT_STEP)
		upper = oldUpper - SOFT_STEP;

	// Modify the lower reading.
	if (lower > oldLower + SOFT_STEP)
		lower = oldLower + SOFT_STEP;
	else if (lower < oldLower - SOFT_STEP)
		lower = oldLower - SOFT_STEP;
}

#include "statgraph.moc"
